from typing import SupportsFloat, SupportsIndex, SupportsInt

from typing_extensions import Buffer, TypeAlias

ReadableBuffer: TypeAlias = Buffer
ConvertibleToInt: TypeAlias = str | ReadableBuffer | SupportsInt | SupportsIndex
ConvertibleToFloat: TypeAlias = str | ReadableBuffer | SupportsFloat | SupportsIndex
ConvertibleToList: TypeAlias = (
    str | tuple | list | ReadableBuffer | SupportsFloat | SupportsIndex
)

__all__ = [n for n in globals()]



"""
Module that contains the ``scifloat`` class.
"""

from typing import NoReturn

from . import ConvertibleToFloat, ConvertibleToList
from .sciround import round_uncertainties, round_value


class scifloat:
    """
    Scientific variant of measurements with corresponding errors.
    Values are rounded based on their errors, which itself are
    rounded by scientific standards.

    Uncertainties are rounded to a given number of significant digits.
    Is the first number either a 1 or a 2 we round to two significant digits.
    Otherwise we round only to one significant digit. In any case it is rounded to
    the next number.

    >>> from scifloatpy.sciround import round_uncertainties,round_value
    >>> round_uncertainties(0.0513)
    0.06

    >>> round_uncertainties(1.2421)
    1.3

    Values are rounded to the number of digits of their error.
    >>> round_uncertainties(0.0102)
    0.011
    >>> round_value(1.36551315)
    1.366

    If the measuremnt has assymmetric errors, the one with the smallest precision, i.e.,
    the one with less digits.
    >>> round_uncertainties([0.0302,0.00891])
    [0.03,0.009]
    >>> round_value(1.36551315)
    1.37

    In such cases the other uncertainty needs also to be modified to the same precision.
    >>> 0.009
    0.01
    (Not implemented yet!)
    """

    def __new__(cls, __x: ConvertibleToFloat, __dx: ConvertibleToList) -> "scifloat":
        return super().__new__(cls)

    def __init__(self, __x: ConvertibleToFloat, __dx: ConvertibleToList) -> NoReturn:
        self.x = float(__x)
        if isinstance(__dx, (int, float)):
            self.dx = [__dx, __dx]
        elif isinstance(__dx, tuple):
            self.dx = list(__dx)
        else:
            self.dx = __dx
        self.rx = self.x
        self.rdx = self.dx.copy()
        self.sx = self.x
        self.sdx = self.dx.copy()

    def __str__(self) -> str:
        if self.sdx[0] == self.sdx[1]:
            return f"{self.sx} \u00b1 {self.sdx[0]}"
        return f"{self.sx} +{self.sdx[0]} | -{self.sdx[1]}"

    def __round__(self, ndigits: None = None) -> "scifloat":
        rdx = round_uncertainties(self.dx.copy())
        if any(list(map(lambda x: "." in x, rdx))):
            ndigits = min(map(lambda x: len(x.split(".")[-1]), rdx))
        else:
            ndigits = -max(map(lambda x: len(x) - len(x.rstrip("0")), rdx))

        rx = round_value(self.x, ndigits)
        self.rx = float(rx)
        self.rdx = list(map(float, rdx))
        self.sx = rx
        self.sdx = rdx
        print(self)
        return self



"""
Module with the two methods to round the scientific float.

Example
-------
>>> round_uncertainties([0.124,0.201])
    [0.13,0.21]

>>> round_value(1.235511,2)
    1.24
"""

import decimal
import math

from . import ConvertibleToFloat, ConvertibleToList

__all__ = [n for n in globals() if n[0] != "_"]


def _float_to_str(__x: ConvertibleToFloat):
    """
    Convert the given float to a string,
    without resorting to scientific notation
    """
    dcx = decimal.Context()
    dcx.prec = 30
    d1 = dcx.create_decimal(repr(__x))
    return format(d1, "f")


def _order_of_magnitude(__x: ConvertibleToFloat):
    if __x == 0:
        return __x
    return math.floor(math.log(__x, 10))


def _significant_numbers(__x: ConvertibleToFloat):
    __s = _float_to_str(__x)
    if __s[0] in ["1", "2"]:
        return 2
    return 1


def _sci_round(__x: ConvertibleToFloat, __sig_num: int):
    __s: str = _float_to_str(__x)
    if len(__s.replace(".", "")) == __sig_num:
        return float(__s)
    if __s == "0":
        return 0
    if __sig_num == 2:
        if not __s[3] == "0":
            return float(__s[:3]) + 0.1
        return float(__s[:3])
    else:
        if not __s[2] == "0":
            return int(__s[:1]) + 1
        return int(__s[:1])


def _number_to_str(__x, __sig_num, __magnitude):
    __s = _float_to_str(__x)
    nmod = __s.replace(".", "")
    nmod = nmod.lstrip("0")
    nmod = nmod.rstrip("0")
    if len(nmod) < __sig_num:
        return __s + "0"
    if len(nmod) > __sig_num:
        return _float_to_str(round(__x, -__magnitude + 1))
    return __s


def _round_uncertainty(__x):
    mag = _order_of_magnitude(__x)
    xflat = __x * 10**-mag
    nsig = _significant_numbers(xflat)
    xround = _sci_round(xflat, nsig)
    xround = xround * 10**mag
    if mag > 0:
        xround = int(xround)
    return _number_to_str(xround, nsig, mag)


def round_uncertainties(__dx: ConvertibleToList) -> list:
    """
    Returns the, by Scientific rules, rounded uncertainties.
    """
    if isinstance(__dx, (int, float)):
        __dx = [__dx, __dx]
    if isinstance(__dx, tuple):
        __dx = list(__dx)
    for i, dx in enumerate(__dx):
        __dx[i] = _round_uncertainty(dx)
    return __dx


def round_value(__x: ConvertibleToFloat, __ndigits: int) -> float:
    """
    Returns the, by Scientific rules, rounded value.

    Examples
    -------
    >>> round_value(1.25,1)
    1.2

    >>> round_value(1.315,2)
    1.32
    """
    # need to remove computation bug with floating points
    if not __ndigits < 0:
        __x = math.floor(__x * 10 ** (__ndigits + 1)) / 10 ** (__ndigits + 1)
        # even rounding, i.e., statistically correct
        return round(__x * 10**__ndigits) / 10**__ndigits
    return round(int(__x), __ndigits)
